#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#include <math.h>
#include <SDL.h>
#include "image.h"
#include "color_bits.h"

void redraw(void);

SDL_Surface *fbsurf;
const int scrx = 640, scry = 480;
float ray_x = 0, ray_y = 0;
uint32_t *offset_map, *nref_map;
unsigned long oxsz, oysz, nxsz, nysz;

#define RAY_COUNT	2000

float subray_x[RAY_COUNT];
float subray_y[RAY_COUNT];

float ray_radius = 0.02;

float frand(float range) {
	return range * (float)rand() / (float)RAND_MAX;
}

float dot(float x1, float y1, float z1, float x2, float y2, float z2) {
	return x1 * x2 + y1 * y2 + z1 * z2;
}

void unpack_vec(uint32_t pixel, float *x, float *y, float *z);
void normalize(float *x, float *y, float *z);

#define SQ(x)	((x) * (x))
#define MIN(a, b)	((a) < (b) ? (a) : (b))
#define MAX(a, b)	((a) > (b) ? (a) : (b))

int main(void) {
	int i, done = 0;
	uint32_t *ptr;
	float lx = -1.0;
	float ly = 1.0;
	float lz = 1.0;
	float vx = 0.0;
	float vy = 0.0;
	float vz = 1.0;


	for(i=0; i<RAY_COUNT; i++) {
		subray_x[i] = frand(ray_radius * 2.0) - ray_radius;
		subray_y[i] = frand(ray_radius * 2.0) - ray_radius;
	}
	
	SDL_Init(SDL_INIT_VIDEO);
	if(!(fbsurf = SDL_SetVideoMode(scrx, scry, 32, SDL_SWSURFACE))) {
		return EXIT_FAILURE;
	}

	if(!(offset_map = load_image("out.png", &oxsz, &oysz))) {
		SDL_Quit();
		return EXIT_FAILURE;
	}

	if(!(nref_map = load_image("normal.png", &nxsz, &nysz))) {
		SDL_Quit();
		return EXIT_FAILURE;
	}

	normalize(&lx, &ly, &lz);
	normalize(&vx, &vy, &vz);
	
	ptr = nref_map;
	for(i=0; i<nxsz*nysz; i++) {
		float nx, ny, nz, hx, hy, hz;
		float n_dot_h, n_dot_l, specular;
		int sint;

		unpack_vec(*ptr, &nx, &ny, &nz);
		ny = -ny;

		normalize(&nx, &ny, &nz);

		hx = (vx + lx) * 0.5;
		hy = (vy + ly) * 0.5;
		hz = (vz + lz) * 0.5;
		
		normalize(&hx, &hy, &hz);

		n_dot_l = dot(nx, ny, nz, lx, ly, lz);

		n_dot_h = dot(nx, ny, nz, hx, hy, hz);
		if(n_dot_h < 0.0) n_dot_h = -n_dot_h;
		specular = pow(n_dot_h, 50.0);

		sint = MIN((int)(specular * 255.0), 255);
		*ptr++ = PACK_COLOR24(sint, sint, sint);
	}

	while(!done) {
		SDL_Event event;
		if(SDL_PollEvent(&event)) {
			switch(event.type) {
			case SDL_KEYDOWN:
				if(event.key.keysym.sym == SDLK_ESCAPE) {
					done = 1;
				}
				break;

			case SDL_QUIT:
				done = 1;
				break;

			case SDL_MOUSEMOTION:
				if(event.motion.state & SDL_BUTTON(1)) {
					ray_x = (float)event.motion.x / (float)scrx;
					ray_y = (float)event.motion.y / (float)scry;
				}
				break;

			case SDL_MOUSEBUTTONDOWN:
				/*if(event.button.button == SDL_BUTTON_RIGHT) {
				}*/
				break;

			default:
				break;
			}
		} else {
			redraw();
		}
	}

	SDL_Quit();
	return 0;
}

void unpack_vec(uint32_t pixel, float *x, float *y, float *z) {
	*x = ((pixel & RED_MASK32) >> RED_SHIFT32) / 128.0f - 1.0f;
	*y = ((pixel & GREEN_MASK32) >> GREEN_SHIFT32) / 128.0f - 1.0f;
	*z = ((pixel & BLUE_MASK32) >> BLUE_SHIFT32) / 128.0f - 1.0f;
}


uint32_t add_color(uint32_t pixel) {
	int r, g, b;
	int ar = 255;
	int ag = 10;
	int ab = 5;

	r = (pixel & RED_MASK32) >> RED_SHIFT32;
	g = (pixel & GREEN_MASK32) >> GREEN_SHIFT32;
	b = (pixel & BLUE_MASK32) >> BLUE_SHIFT32;

	return PACK_COLOR24(MIN(r + ar, 255), MIN(g + ag, 255), MIN(b + ab, 255));
}

void redraw(void) {
	int i;
	int sx, sy, tx, ty;
	float off_x, off_y, off_z;
	uint32_t *pix;
	
	/*SDL_FillRect(fbsurf, 0, 0);*/

	if(SDL_MUSTLOCK(fbsurf)) SDL_LockSurface(fbsurf);

	for(i=0; i<scrx*scry; i++) {
		((uint32_t*)fbsurf->pixels)[i] = nref_map[i];
	}

	for(i=0; i<RAY_COUNT; i++) {
		tx = (ray_x + subray_x[i]) * oxsz;
		ty = (ray_y + subray_y[i]) * oysz;

		if(tx < 0 || tx >= oxsz || ty < 0 || ty >= oysz) {
			continue;
		}
		
		unpack_vec(offset_map[ty * oxsz + tx], &off_x, &off_y, &off_z);

		sx = (ray_x + off_x * 0.6) * scrx;
		sy = (ray_y + off_y * 0.6) * scry;

		if(sx >= 0 && sx < scrx - 1 && sy >= 0 && sy < scry - 1) {
			pix = (uint32_t*)fbsurf->pixels + sy * scrx + sx;
			*pix = add_color(*pix);
			if(sx > 0) *(pix - 1) = add_color(*(pix - 1));
			if(sx < scrx - 1) *(pix + 1) = add_color(*(pix + 1));
			if(sy > 0) *(pix - scrx) = add_color(*(pix - scrx));
			if(sy < scry - 1) *(pix + scrx) = add_color(*(pix + scrx));
		}
	}

	if(SDL_MUSTLOCK(fbsurf)) SDL_UnlockSurface(fbsurf);

	SDL_Flip(fbsurf);
}

void normalize(float *x, float *y, float *z) {
	float len = (float)sqrt(SQ(*x) + SQ(*y) + SQ(*z));
	*x /= len;
	*y /= len;
	*z /= len;
}
